import numpy as np
import thread
import xml.etree.cElementTree as ET
import subprocess
import time
from scipy.optimize import basinhopping
from datetime import datetime
from sympy import Point3D, Line3D
import matplotlib.pyplot as plt
import os

model_nr = 'du02'

method = 'TNC'
# method = 'LBFGSB'

# number = 1
# number = 2
# number = 3
# number = 4
number = 5

global counter_AP
global counter_IE
global counter_VV
global counter_run

counter_AP = 0
counter_IE = 0
counter_VV = 0
counter_run = 0
thread_count = 0

# ----------------------------------------------------------------------------------------------------------------------
# --- SET DIRECTORIES AND FILEPATHS ------------------------------------------------------------------------------------
dir_robot_data = r'./Robot_data'
dir_template_feb = r'./Templates_{}_{}_{}'.format(model_nr, method, number)
dir_run_feb = r'./Runfolder_{}_{}_{}'.format(model_nr, method, number)
dir_run_feb_output = r'{}/Output'.format(dir_run_feb)

# file names
feb_name_imaging_posture = '{}'.format(model_nr) 
feb_name_PS_F_AL = '{}_PS_F_AL'.format(model_nr)    # Prestretch, flexion and axial load
feb_name_VV = '{}_VV'.format(model_nr)
feb_name_IE = '{}_IE'.format(model_nr)
feb_name_AP = '{}_AP'.format(model_nr)
file_names_list = [feb_name_PS_F_AL, feb_name_VV, feb_name_IE, feb_name_AP]

log_file_PS_F_AL = ('{}/{}.log'.format(dir_run_feb, feb_name_PS_F_AL))
log_file_VV = ('{}/{}.log'.format(dir_run_feb, feb_name_VV))
log_file_AP = ('{}/{}.log'.format(dir_run_feb, feb_name_AP))
log_file_IE = ('{}/{}.log'.format(dir_run_feb, feb_name_IE))

min_load_VV = -5000
max_load_VV = 5000
min_load_AP = -80
max_load_AP = 60
min_load_IE = -4000
max_load_IE = 4000

########################################################################################################################
# ------------------------------------------------- LOAD ROBOT DATA ----------------------------------------------------
# Load robot data look up tables at 0 degrees knee flexion (Angles in degrees)
robot_AP_forces = np.load(r'{}/{}/AP_forces.npy'.format(dir_robot_data, model_nr))
robot_AP_positions = np.load(r'{}/{}/AP_position.npy'.format(dir_robot_data, model_nr))
robot_IE_moments = np.load(r'{}/{}/IE_moments.npy'.format(dir_robot_data, model_nr))
robot_IE_angles = np.load(r'{}/{}/IE_angles.npy'.format(dir_robot_data, model_nr))
robot_VV_moments = np.load(r'{}/{}/VV_moments.npy'.format(dir_robot_data, model_nr))
robot_VV_angles = np.load(r'{}/{}/VV_angles.npy'.format(dir_robot_data, model_nr))

# Min & Max values Robot data 
# min_robot_AP = -22.0790
# max_robot_AP = 6.3925
# min_robot_IE = np.deg2rad(-29.9830)
# max_robot_IE = np.deg2rad(24.3440)
# min_robot_VV = np.deg2rad(-11.2230)
# max_robot_VV = np.deg2rad(-0.0954)
# normalization_AP = max([abs(min_robot_AP), abs(max_robot_AP)])
# normalization_IE = max([abs(min_robot_IE), abs(max_robot_IE)])
# normalization_VV = max([abs(min_robot_VV), abs(max_robot_VV)])


def find_closest_value_in_array(array, value):
    closest_value = array[np.abs(array - value).argmin()]
    index = list(array).index(closest_value)
    return index


def get_robot_position_AP(simulated_force_ap):
    if isinstance(simulated_force_ap, int):
        index_closest_value_AP_force = find_closest_value_in_array(robot_AP_forces, simulated_force_ap)
        AP_robot_positions = robot_AP_positions[0][index_closest_value_AP_force]
    else:
        AP_robot_positions = []
        for i in range(0, len(simulated_force_ap)):
            index_closest_value_AP_force = find_closest_value_in_array(robot_AP_forces, simulated_force_ap[i])
            AP_robot_positions.append(robot_AP_positions[0][index_closest_value_AP_force])

    return AP_robot_positions


def get_robot_angle_IE(simulated_moment_ie):
    if isinstance(simulated_moment_ie, int):
        index_closest_value_IE_moment = find_closest_value_in_array(robot_IE_moments, simulated_moment_ie)
        IE_robot_angles = robot_IE_angles[0][index_closest_value_IE_moment]
    else:
        IE_robot_angles = []
        for i in range(0, len(simulated_moment_ie)):
            index_closest_value_IE_moment = find_closest_value_in_array(robot_IE_moments, simulated_moment_ie[i])
            IE_robot_angles.append(robot_IE_angles[0][index_closest_value_IE_moment])

    return IE_robot_angles


def get_robot_angle_VV(simulated_moment_vv):
    if isinstance(simulated_moment_vv, int):
        index_closest_value_VV_moment = find_closest_value_in_array(robot_VV_moments, simulated_moment_vv)
        VV_robot_angles = robot_VV_angles[0][index_closest_value_VV_moment]
    else:
        VV_robot_angles = []
        for i in range(0, len(simulated_moment_vv)):
            index_closest_value_VV_moment = find_closest_value_in_array(robot_VV_moments, simulated_moment_vv[i])
            VV_robot_angles.append(robot_VV_angles[0][index_closest_value_VV_moment])

    return VV_robot_angles

max_robot_AP = get_robot_position_AP(max_load_AP)
min_robot_AP = get_robot_position_AP(min_load_AP)
max_robot_IE = get_robot_angle_IE(max_load_IE)
min_robot_IE = get_robot_angle_IE(min_load_IE)
max_robot_VV = get_robot_angle_VV(max_load_VV)
min_robot_VV = get_robot_angle_VV(min_load_VV)
normalization_AP = max([abs(min_robot_AP), abs(max_robot_AP)])
normalization_IE = max([abs(min_robot_IE), abs(max_robot_IE)])
normalization_VV = max([abs(min_robot_VV), abs(max_robot_VV)])

########################################################################################################################
# ------------------------------------------------- FUNCTIONS ----------------------------------------------------------


def vec_mag(x):
    mag_x = np.sqrt(x[0] ** 2 + x[1] ** 2 + x[2] ** 2)

    return mag_x


def unit_vec(x, y):
    x = np.array(x)
    y = np.array(y)
    xy = y - x
    mag_xy = np.sqrt(xy[0] ** 2 + xy[1] ** 2 + xy[2] ** 2)
    u_xy = xy / mag_xy

    return mag_xy, u_xy


def set_load_values_in_lc(tree_knee, root_knee, max_load, min_load, simulation):
    if simulation == 'VV':
        lc = '9'
        max_time = 1.4
        min_time = 2.2
    elif simulation == 'AP':
        lc = '10'
        max_time = 1.4
        min_time = 2.2
    elif simulation == 'IE':
        lc = '7'
        max_time = 1.4
        min_time = 2.2

    for child in root_knee.findall('LoadData/loadcurve'):
        if child.attrib['id'] == '{}'.format(lc):
            max = '{},{}'.format(max_time, max_load)
            min = '{},{}'.format(min_time, min_load)
            child[2].text = max
            child[3].text = min

    return tree_knee


def set_prestretch_values(feb_file_name, tree_knee_feb, root_knee_feb, stretch_values, YMvalues):
    for child in root_knee_feb.findall('LoadData/loadcurve'):   
        if child.attrib['id'] == '4':                       # LCL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[0])
            step2 = '4,{}'.format(stretch_values[0])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2
             
        if child.attrib['id'] == '3':                       # MCL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[1])
            step2 = '4,{}'.format(stretch_values[1])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2
            
        if child.attrib['id'] == '1':                       # ACL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[2])
            step2 = '4,{}'.format(stretch_values[2])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2

        if child.attrib['id'] == '2':                       # PCL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[3])
            step2 = '4,{}'.format(stretch_values[3])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2
            
    for child in root_knee_feb.findall('Material/material'):
        if child.attrib['name'] == 'ACL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[0]))

        if child.attrib['name'] == 'PCL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[1]))

        if child.attrib['name'] == 'MCL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[2]))

        if child.attrib['name'] == 'LCL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[3]))
            

    tree_knee_feb.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_PS_F_AL), encoding="ISO-8859-1",
                        xml_declaration=True)


def run_feb(feb_file_name):
    global thread_count
    print('{} run starting'.format(feb_file_name))

    subprocess.call(["./FEBio-3.0.0/bin/febio3", '-i', "{}/{}.feb".format(dir_run_feb, feb_file_name), '-dump', ])

    print('{} run finished'.format(feb_file_name))
    thread_count += 1


def run_feb_from_restart(feb_file_name):
    global thread_count
    print('{} run starting'.format(feb_file_name))

    subprocess.call(["./FEBio-3.0.0/bin/febio3", '-r', "{}/{}.feb".format(dir_run_feb, feb_file_name)])

    print('{} run finished'.format(feb_file_name))
    thread_count += 1


def read_logfile(log_file):
    # Most distal point on the posterior surface midway between medial & lateral condyles                # 523
    # Medial most posterior point on the femoral condyle                                                 # 771
    # Lateral most posterior point on the femoral condyle                                                # 899
    # Medial center of tibial plateau                                                                    # 15242
    # Lateral center of tibial plateau                                                                   # 15349
    # Medial point of intercondylar eminence                                                             # 15589
    # Lateral point of intercondylar eminence                                                            # 15595
    # Closest node to Center of distal intramedullary canal                                              # 12608
    # Femur head anterior    (No DU02 ALM digitized data, used same node number as in previous models)   # 7086
    # Femur head posterior   (No DU02 ALM digitized data, used same node number as in previous models)   # 7070
    # Node opposite to closest node to Center of distal intramedullary canal                             # 12616
    Node_7070 = []
    Node_7086 = []
    Node_523 = []
    Node_12608 = []
    Node_12616 = []
    Node_15349 = []
    Node_15589 = []
    Node_15595 = []
    Node_899 = []
    Node_771 = []
    Node_15242 = []
    RCJ1 = []
    RCJ2 = []
    RCJ3 = []
    converged_times = []

    with open(log_file, 'r') as inf:
        input_data = inf.readlines()
        for ii in range(0, len(input_data)):
            line = input_data[ii]

            if line.startswith('7070 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_7070.append(xyz)

            elif line.startswith('7086 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_7086.append(xyz)

            elif line.startswith('523 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_523.append(xyz)

            elif line.startswith('12608 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_12608.append(xyz)

            elif line.startswith('12616 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_12616.append(xyz)

            elif line.startswith('15349 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_15349.append(xyz)

            elif line.startswith('15589 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_15589.append(xyz)

            elif line.startswith('15595 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_15595.append(xyz)

            elif line.startswith('899 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_899.append(xyz)

            elif line.startswith('771 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_771.append(xyz)

            elif line.startswith('15242 '):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                Node_15242.append(xyz)

            elif line.startswith("Data = RCFx;RCFy;RCFz;RCMx;RCMy;RCMz\n"):
                xyz1 = [float(x) for x in input_data[ii + 1].split(',')][1:]
                RCJ1.append(xyz1)
                xyz2 = [float(x) for x in input_data[ii + 2].split(',')][1:]
                RCJ2.append(xyz2)
                xyz3 = [float(x) for x in input_data[ii + 3].split(',')][1:]
                RCJ3.append(xyz3)

            elif line.startswith("------- converged at time :"):
                A = input_data[ii][27:]
                B = A.replace('\n', '')
                converged_times.append(float(B))
                
            elif line.startswith(" Elapsed time :"):
                A = input_data[ii][15:]
                B = A.replace('\n', '')
                elapsed_time = B

    RCJ1 = np.asarray(RCJ1)
    RCJ2 = np.asarray(RCJ2)
    RCJ3 = np.asarray(RCJ3)

    number_of_steps_converged = len(Node_15242)
    ALM_Nodal_nrs = [7070, 7086, 523, 12608, 12616, 15349, 15589, 15595, 899, 771, 15242]

    ALM_Nodal_Coords = []
    for r in range(0, number_of_steps_converged):
        ALM_Nodal_Coords.append(np.asarray(np.zeros((len(ALM_Nodal_nrs), 4))))

    for i in range(0, number_of_steps_converged):
        ALM_Nodal_Coords[i][0, :] = Node_7070[i]
        ALM_Nodal_Coords[i][1, :] = Node_7086[i]
        ALM_Nodal_Coords[i][2, :] = Node_523[i]
        ALM_Nodal_Coords[i][3, :] = Node_12608[i]
        ALM_Nodal_Coords[i][4, :] = Node_12616[i]
        ALM_Nodal_Coords[i][5, :] = Node_15349[i]
        ALM_Nodal_Coords[i][6, :] = Node_15589[i]
        ALM_Nodal_Coords[i][7, :] = Node_15595[i]
        ALM_Nodal_Coords[i][8, :] = Node_899[i]
        ALM_Nodal_Coords[i][9, :] = Node_771[i]
        ALM_Nodal_Coords[i][10, :] = Node_15242[i]

    return ALM_Nodal_Coords, RCJ1, RCJ2, RCJ3, converged_times
    

def Knee_ACS(ALM_nodal_Coords, step):
    dist_between_femcondyles = np.array(ALM_nodal_Coords[step][2, 1:])  # 523
    med_post_fem = np.array(ALM_nodal_Coords[step][9, 1:])  # 771
    lat_post_fem = np.array(ALM_nodal_Coords[step][8, 1:])  # 899
    med_center_tibplateau = np.array(ALM_nodal_Coords[step][10, 1:])  # 15242
    lat_center_tibplateau = np.array(ALM_nodal_Coords[step][5, 1:])  # 15349
    med_intercond_eminence = np.array(ALM_nodal_Coords[step][6, 1:])  # 15589
    lat_intercond_eminence = np.array(ALM_nodal_Coords[step][7, 1:])  # 15595
    center_dist_canal_closestnode = np.array(ALM_nodal_Coords[step][3, 1:])  # 12608
    opposite_center_dist_canal_closestnode = np.array(ALM_nodal_Coords[step][4, 1:])  # 12616
    fem_head_anterior = np.array(ALM_nodal_Coords[step][1, 1:])  # 7086
    fem_head_posterior = np.array(ALM_nodal_Coords[step][0, 1:])  # 7070

    # Calculate the center of the femoral head.
    center_fem_head = (fem_head_anterior + fem_head_posterior) / 2

    # Calculate location of center_dist_canal.
    # Use closest mesh node and opposite node to closest node, find point on line between these points using the factor.
    factor = 0.4034134501615193047  # as found in DU02_ACS_alignment_model2.py (percentage on line between node 1 and node 2)
    center_dist_canal = center_dist_canal_closestnode + factor * (
                opposite_center_dist_canal_closestnode - center_dist_canal_closestnode)

    #  Femur local coordinate system ***********************************************************************************
    # Calculate femur origin: on the femoral mechanical axis (Z) at the most distal bony point midway between condyles.
    p1_Z = Point3D(dist_between_femcondyles)
    p2_Z = Point3D(center_fem_head)
    p3_dbf = Point3D(dist_between_femcondyles)
    l1 = Line3D(p1_Z, p2_Z)
    projection_fem = l1.projection(p3_dbf)
    eval_fem = projection_fem.evalf()
    fem_origin = np.array([float(eval_fem.x), float(eval_fem.y), float(eval_fem.z)], )

    # Mechanical axis (Z): from center of femoral head to most distal point on posterior surface between condyles.
    fem_Z_mag, fem_Z_unit = unit_vec(dist_between_femcondyles, center_fem_head)
    K_unit = fem_Z_unit

    # Line used to calculate Y axis later: between lateral and medial most posterior point on femoral condyles.
    fem_calc_line_mag, fem_calc_line_unit = unit_vec(med_post_fem, lat_post_fem)

    # Y axis: cross product between mechanical axis Z and line calculated above.
    fem_Y_unit = np.cross(fem_Z_unit, fem_calc_line_unit) / vec_mag(np.cross(fem_Z_unit, fem_calc_line_unit))
    J_unit = fem_Y_unit

    # X axis: cross product between Y and Z unit vectors (finish the right-handed coordinate system)
    fem_X_unit = np.cross(fem_Y_unit, fem_Z_unit) / vec_mag(np.cross(fem_Y_unit, fem_Z_unit))
    I_unit = fem_X_unit

    fem_base_vecs = [I_unit, J_unit, K_unit]

    #  Tibia local coordinate system *******************************************************************************
    # Mechanical axis z: midway between two intercondylar eminences and center of ankle distally
    # (center distal intramedullary canal)
    midway_intercond_eminence = (lat_intercond_eminence + med_intercond_eminence) / 2
    tib_z_mag, tib_z_unit = unit_vec(center_dist_canal, midway_intercond_eminence)
    k_unit = tib_z_unit

    # Calculate the tibial origin
    p1_z = Point3D(center_dist_canal)
    p2_z = Point3D(midway_intercond_eminence)
    p3_mid_eminence = Point3D(midway_intercond_eminence)
    l1 = Line3D(p1_z, p2_z)
    projection_tib = l1.projection(p3_mid_eminence)
    eval_tib = projection_tib.evalf()
    tib_origin = np.array([float(eval_tib.x), float(eval_tib.y), float(eval_tib.z)], )

    # Line used to calculate y later: between center of the medial and lateral tibial plateau
    tib_calc_line_mag, tib_calc_line_unit = unit_vec(med_center_tibplateau, lat_center_tibplateau)

    # y axis: cross product between mechanical axis z and line calculated above.
    tib_y_unit = np.cross(tib_z_unit, tib_calc_line_unit) / vec_mag(np.cross(tib_z_unit, tib_calc_line_unit))
    j_unit = tib_y_unit

    # x axis: cross product between y and z unit vectors (finish the right-handed coordinate system)
    tib_x_unit = np.cross(tib_y_unit, tib_z_unit) / vec_mag(np.cross(tib_y_unit, tib_z_unit))
    i_unit = tib_x_unit

    tib_base_vecs = [i_unit, j_unit, k_unit]

    return [fem_base_vecs, tib_base_vecs, fem_origin, tib_origin]

    
def TF_joint_transformation(femur_base_vecs, tibia_base_vecs, femur_origin, tibia_origin):
    # ONLY FOR RIGHT KNEES!
    [I, J, K] = femur_base_vecs
    [i, j, k] = tibia_base_vecs

    e1 = I
    e3 = k
    e2 = np.cross(e3, e1) / vec_mag(np.cross(e3, e1))  # Floating axis

    alpha = np.arcsin(np.dot(-e2, K))  # Flexion (+) - Extension (-)
    flexion = alpha
    beta = np.arccos(np.dot(I, k))
    adduction = beta - (np.pi / 2)  # Abduction (+)(Valgus) - Adduction (-) (Varus)
    gamma = np.arcsin(np.dot(-e2, i))  # External rotation (+) - Internal (-)
    ext_rot = gamma

    # H: translation between femur and tibia vector, directed from femoral origin to tibial origin.
    H = tibia_origin - femur_origin
    q1 = np.dot(H, e1)  # Medial (-) - lateral (+) tibial thrust
    ML_trans = q1
    q2 = np.dot(H, e2)  # Anterior (+) - posterior (-) tibial drawer
    AP_trans = q2
    q3 = np.dot(-H,
                e3)  # Joint distraction (+) - compression (-): height of the femoral origin above the tibial transverse plane.
    Joint_distraction = q3

    return [flexion, adduction, ext_rot, ML_trans, AP_trans, Joint_distraction]


def get_kinematics_and_loads(log_file):
    list_of_all_nodes, rcj1, rcj2, rcj3, converged_times = read_logfile(log_file)

    all_steps_ACS = []
    all_steps_rot_trans = []
    rot_trans_from_init = []

    for n in range(0, len(converged_times)):
        init_ACS = Knee_ACS(list_of_all_nodes, step=0)
        init_rot_trans = TF_joint_transformation(init_ACS[0], init_ACS[1], init_ACS[2], init_ACS[3])

        all_steps_ACS.append(Knee_ACS(list_of_all_nodes, step=n))
        all_steps_rot_trans.append(TF_joint_transformation(all_steps_ACS[n][0], all_steps_ACS[n][1], all_steps_ACS[n][2], all_steps_ACS[n][3]))
        rot_trans_from_init.append(np.array(all_steps_rot_trans[n]) - np.array(init_rot_trans))

    position = np.transpose(all_steps_rot_trans)

    return position, rcj1, rcj2, rcj3, converged_times


def loads_applied(converged_times, min_load, max_load):
    applied_loads = []
    for i in range(0, len(converged_times)):
        if converged_times[i] <= 1.0:
            load = 0

        elif converged_times[i] > 1.0 and converged_times[i] <= 1.4:
            b = max_load - ((max_load - 0) / (1.4 - 1.0)) * 1.4
            load = (((max_load - 0) / (1.4 - 1.0)) * converged_times[i]) + b

        elif converged_times[i] > 1.4 and converged_times[i] <= 2.2:
            b = min_load - (((min_load - max_load) / (2.2 - 1.4)) * 2.2)
            load = (((min_load - max_load) / (2.2 - 1.4)) * converged_times[i]) + b

        applied_loads.append(load)

    return applied_loads


def calc_error_vv(robot_data, simulated_data, converged_timestep, normalization_value):
    global counter_VV
    times = [1.0, 1.3, 1.6, 1.9, 2.2]

    if converged_timestep >= times[4]:
        error = np.sqrt(
            (np.array([x ** 2 for x in np.subtract(simulated_data, robot_data)]).sum()) / len(simulated_data))
        error_norm = error / normalization_value
        counter_VV = 0

    else:
        if converged_timestep >= times[3]:
            error_norm = 3.125
        elif converged_timestep >= times[2]:
            error_norm = 3.150
        elif converged_timestep >= times[1]:
            error_norm = 3.175
        elif converged_timestep >= times[0]:
            error_norm = 3.200

        counter_VV += 1
        if counter_VV >= 100:
            counter_VV_100 = 100
            error_norm = error_norm * (1 + (0.001 * (counter_VV_100 - 1)))
        else:
            error_norm = error_norm * (1 + (0.001 * (counter_VV - 1)))

    return error_norm


def calc_error_ap(robot_data, simulated_data, converged_timestep, normalization_value):
    global counter_AP
    times = [1.0, 1.3, 1.6, 1.9, 2.2]

    if converged_timestep >= times[4]:
        error = np.sqrt(
            (np.array([x ** 2 for x in np.subtract(simulated_data, robot_data)]).sum()) / len(simulated_data))
        error_norm = error / normalization_value
        counter_AP = 0

    else:
        if converged_timestep >= times[3]:
            error_norm = 3.125
        elif converged_timestep >= times[2]:
            error_norm = 3.150
        elif converged_timestep >= times[1]:
            error_norm = 3.175
        elif converged_timestep >= times[0]:
            error_norm = 3.200

        counter_AP += 1
        if counter_AP >= 100:
            counter_AP_100 = 100
            error_norm = error_norm * (1 + (0.001 * (counter_AP_100 - 1)))
        else:
            error_norm = error_norm * (1 + (0.001 * (counter_AP - 1)))

    return error_norm


def calc_error_ie(robot_data, simulated_data, converged_timestep, normalization_value):
    global counter_IE
    times = [1.0, 1.3, 1.6, 1.9, 2.2]

    if converged_timestep >= times[4]:
        error = np.sqrt(
            (np.array([x ** 2 for x in np.subtract(simulated_data, robot_data)]).sum()) / len(simulated_data))
        error_norm = error / normalization_value
        counter_IE = 0

    else:
        if converged_timestep >= times[3]:
            error_norm = 3.125
        elif converged_timestep >= times[2]:
            error_norm = 3.150
        elif converged_timestep >= times[1]:
            error_norm = 3.175
        elif converged_timestep >= times[0]:
            error_norm = 3.200

        counter_IE += 1
        if counter_IE >= 100:
            counter_IE_100 = 100
            error_norm = error_norm * (1 + (0.001 * (counter_IE_100 - 1)))
        else:
            error_norm = error_norm * (1 + (0.001 * (counter_IE - 1)))

    return error_norm
    

def flexion_application_check(log_file):
    converged_times = []

    with open(log_file, 'r') as inf:
        input_data = inf.readlines()
        for ii in range(0, len(input_data)):
            line = input_data[ii]

            if line.startswith("------- converged at time :"):
                A = input_data[ii][27:]
                B = A.replace('\n', '')
                converged_times.append(float(B))

    if converged_times[-1] == 1.0:
        flexion_done = 1
    else:
        flexion_done = 0

    last_convergedtime = converged_times[-1]

    return flexion_done, last_convergedtime


def check_if_combination_already_ran(PS, YMV):
    AP_error_prev = np.nan
    VV_error_prev = np.nan
    IE_error_prev = np.nan
    already_ran = 0
    # prestretch_values = [LCL, MCL, ACL, PCL]

    if counter_run > 0:
        prestretch_values_file = '{}/prestretch_values_LCL_MCL_ACL_PCL_rounded.txt'.format(dir_run_feb_output)
        total_errors_file = '{}/{}_normalized_errors.txt'.format(dir_run_feb_output, model_nr)

        with open(prestretch_values_file, 'r') as inf:
            input_data = inf.readlines()

            values = []
            for ii in range(0, len(input_data)):
                values.append([x for x in input_data[ii].split(' ')])

            inf.close()

        ACL = []
        PCL = []
        MCL = []
        LCL = []
        ACL_Y = []
        PCL_Y = []
        MCL_Y = []
        LCL_Y = []
        
        for i in range(0, len(values)):
            ACL.append(float(values[i][4]))
            PCL.append(float(values[i][5]))
            MCL.append(float(values[i][3]))
            LCL.append(float(values[i][2]))
            ACL_Y.append(float(values[i][6]))
            PCL_Y.append(float(values[i][7]))
            MCL_Y.append(float(values[i][8]))
            LCL_Y.append(float(values[i][9]))

        ran = []
        for n in range(0, len(ACL)):
            if ACL[n] == PS[2] and PCL[n] == PS[3] and MCL[n] == PS[1] and LCL[n] == PS[0] and ACL_Y[n] == YMV[0] and PCL_Y[n] == YMV[1] and MCL_Y[n] == YMV[2] and LCL_Y[n] == YMV[3]:
                ran.append(1)
            else:
                ran.append(0)

        if (1 in ran):
            already_ran = 1

            with open(total_errors_file, 'r') as inf:
                input_data = inf.readlines()

                values = []
                for ii in range(0, len(input_data)):
                    values.append([x for x in input_data[ii].split(' ')])

                inf.close()

            AP_error = []
            VV_error = []
            IE_error = []

            for k in range(0, len(values)):
                AP_error.append(float(values[k][5]))
                VV_error.append(float(values[k][3]))
                IE_error.append(float(values[k][7]))

            for m in range(0, len(ran)):
                if ran[m] == 1:
                    AP_error_prev = AP_error[m]
                    VV_error_prev = VV_error[m]
                    IE_error_prev = IE_error[m]

        else:
            already_ran = 0

    return already_ran, AP_error_prev, VV_error_prev, IE_error_prev



########################################################################################################################
# ----------------------------------------------END OF FUNCITONS -------------------------------------------------------
########################################################################################################################

# ----------------------------------------------------------------------------------------------------------------------
# --- SET APPLIED LOAD LOADCURVES IN TEMPLATE SIMULATION FILES ---------------------------------------------------------
feb_file_VV = r'{}/{}.feb'.format(dir_template_feb, feb_name_VV)
tree_knee_VV = ET.parse(r'{}'.format(feb_file_VV))
root_knee_VV = tree_knee_VV.getroot()

feb_file_AP = r'{}/{}.feb'.format(dir_template_feb, feb_name_AP)
tree_knee_AP = ET.parse(r'{}'.format(feb_file_AP))
root_knee_AP = tree_knee_AP.getroot()

feb_file_IE = r'{}/{}.feb'.format(dir_template_feb, feb_name_IE)
tree_knee_IE = ET.parse(r'{}'.format(feb_file_IE))
root_knee_IE = tree_knee_IE.getroot()

tree_knee_VV_updated = set_load_values_in_lc(tree_knee_VV, root_knee_VV, max_load_VV, min_load_VV, 'VV')
tree_knee_VV_updated.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_VV), encoding="ISO-8859-1", xml_declaration=True)

tree_knee_AP_updated = set_load_values_in_lc(tree_knee_AP, root_knee_AP, max_load_AP, min_load_AP, 'AP')
tree_knee_AP_updated.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_AP), encoding="ISO-8859-1", xml_declaration=True)

tree_knee_IE_updated = set_load_values_in_lc(tree_knee_IE, root_knee_IE, max_load_IE, min_load_IE, 'IE')
tree_knee_IE_updated.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_IE), encoding="ISO-8859-1", xml_declaration=True)

# ----------------------------------------------------------------------------------------------------------------------
# --- OBJECTIVE FUNCTION -----------------------------------------------------------------------------------------------
def objective_function(toOptimize):
    global thread_count
    global counter_run
    global file_names_list
    global total_error
    global VV_norm_error
    global AP_norm_error
    global IE_norm_error
    
    PS_values = toOptimize[0:4]     # PS_values = [LCL_init, MCL_init, ACL_init, PCL_init]
    
    ACL_YMvalues = [73, 83, 93, 103, 113, 123, 133, 143, 153, 163, 173]
    PCL_YMvalues = [118, 128, 138, 148, 158, 168, 178, 188, 198, 208, 218]
    MCL_YMvalues = [174, 184, 194, 204, 214, 224, 234, 244, 254, 264, 274]
    LCL_YMvalues = [230, 240, 250, 260, 270, 280, 290, 300, 310, 320, 330]
    
    # YMVals = toOptimize[4:8]        # ACL, PCL, MCL, LCL
    YMVals = [ACL_YMvalues[int(round(toOptimize[4]*10))], PCL_YMvalues[int(round(toOptimize[5]*10))], MCL_YMvalues[int(round(toOptimize[6]*10))], LCL_YMvalues[int(round(toOptimize[7]*10))]]
   
    ACL_PS = PS_values[2]
    PCL_PS = PS_values[3]
    MCL_PS = PS_values[1]
    LCL_PS = PS_values[0]

    ACL_PS_r = round(float(ACL_PS), 3)
    PCL_PS_r = round(float(PCL_PS), 3)
    MCL_PS_r = round(float(MCL_PS), 3)
    LCL_PS_r = round(float(LCL_PS), 3)

    LCL = float(LCL_PS_r)
    MCL = float(MCL_PS_r)
    ACL = float(ACL_PS_r)
    PCL = float(PCL_PS_r)

    prestretch_values = [LCL, MCL, ACL, PCL]
    
    ACL_YM = YMVals[0]
    PCL_YM = YMVals[1]
    MCL_YM = YMVals[2]
    LCL_YM = YMVals[3]
    
    ACL_YM_r = YMVals[0]
    PCL_YM_r = YMVals[1]
    MCL_YM_r = YMVals[2]
    LCL_YM_r = YMVals[3]
    
    ACL_Y = float(ACL_YM_r)
    PCL_Y = float(PCL_YM_r)
    MCL_Y = float(MCL_YM_r)
    LCL_Y = float(LCL_YM_r)
    
    YMValues = [ACL_Y, PCL_Y, MCL_Y, LCL_Y] 

    already_ran, previousAPerror, previousVVerror, previousIEerror = check_if_combination_already_ran(prestretch_values, YMValues)

    if already_ran == 1:
        VV_norm_error = previousVVerror
        AP_norm_error = previousAPerror
        IE_norm_error = previousIEerror
        total_error = (VV_norm_error + AP_norm_error + IE_norm_error) / 3

    elif already_ran == 0:
        # DELETE .log, .dmp, .xplt files BEFORE NEXT RUN------------------------------------------------------------
        if counter_run > 0:
            for i in range(0, len(file_names_list)):
                if os.path.isfile('{}/{}.log'.format(dir_run_feb, file_names_list[i])) == True:
                    os.remove('{}/{}.log'.format(dir_run_feb, file_names_list[i]))
                if os.path.isfile('{}/{}.xplt'.format(dir_run_feb, file_names_list[i])) == True:
                    os.remove('{}/{}.xplt'.format(dir_run_feb, file_names_list[i]))
                if os.path.isfile('{}/{}.dmp'.format(dir_run_feb, file_names_list[i])) == True:
                    os.remove('{}/{}.dmp'.format(dir_run_feb, file_names_list[i]))
            if os.path.isfile('{}/{}.feb'.format(dir_run_feb, feb_name_PS_F_AL)) == True:
                os.remove('{}/{}.feb'.format(dir_run_feb, feb_name_PS_F_AL))

        counter_run += 1

        # SET PRESTRETCH IN IMAGING POSTURE FEB FILE & RUN ---------------------------------------------------------
        feb_file = r'{}/{}.feb'.format(dir_template_feb, feb_name_imaging_posture)
        tree_knee = ET.parse(r'{}'.format(feb_file))
        root_knee = tree_knee.getroot()
        set_prestretch_values(feb_name_imaging_posture, tree_knee, root_knee, prestretch_values, YMValues)

        run_feb(feb_name_PS_F_AL)
        flexion_position_PS_F_AL, fl_loads_rcj1_PS_F_AL, fl_loads_rcj2_PS_F_AL, fl_loads_rcj3_PS_F_AL, fl_converged_times_PS_F_AL = get_kinematics_and_loads(log_file_PS_F_AL)
        np.savetxt('{}/Position_after_flexion.txt'.format(dir_run_feb_output), flexion_position_PS_F_AL)
            
        if len(fl_converged_times_PS_F_AL) == 0:
            flexion_succes = 0
        else:
            flexion_succes, last_convergedtime = flexion_application_check(log_file_PS_F_AL)

        if flexion_succes == 1:
            with open('{}/flexion_angle_PS_F_AL.txt'.format(dir_run_feb_output), 'a') as pres1:
                pres1.write('flexion angle: {} Time: {} \n'.format(str(flexion_position_PS_F_AL[0, -1]), datetime.now().strftime("%d/%m/%Y %H:%M:%S")))
            thread_count = 0
            num_method = 3

            thread.start_new_thread(run_feb_from_restart, ('{}'.format(feb_name_VV),))
            thread.start_new_thread(run_feb_from_restart, ('{}'.format(feb_name_AP),))
            thread.start_new_thread(run_feb_from_restart, ('{}'.format(feb_name_IE),))

            while thread_count < num_method:
                print(thread_count)
                time.sleep(1)

            VV_position, VV_loads_rcj1, VV_loads_rcj2, VV_loads_rcj3, VV_converged_times = get_kinematics_and_loads(
                log_file_VV)                    
            AP_position, AP_loads_rcj1, AP_loads_rcj2, AP_loads_rcj3, AP_converged_times = get_kinematics_and_loads(
                log_file_AP)
            IE_position, IE_loads_rcj1, IE_loads_rcj2, IE_loads_rcj3, IE_converged_times = get_kinematics_and_loads(
                log_file_IE)

            if len(VV_converged_times) == 0:
                VV_check = 0
            else:
                VV_check = 1
                
            if len(AP_converged_times) == 0:
                AP_check = 0
            else:
                AP_check = 1
                
            if len(IE_converged_times) == 0:
                IE_check = 0
            else:
                IE_check = 1

            # CALCULATE VV ERROR ---------------- Position: 0: FE, 1: VV, 2: IE, 3: ML, 4: AP, 5: SI ---------------
            if VV_check == 1:
                if len(VV_converged_times) > 0:
                    VV_converged_timestep = VV_converged_times[-1]
                    VV_simulated_loads = loads_applied(VV_converged_times, min_load_VV, max_load_VV)
                    VV_simulated_angles = VV_position[1, :]
                    VV_robot_angles = get_robot_angle_VV(VV_simulated_loads)
                    VV_norm_error = calc_error_vv(VV_robot_angles, VV_simulated_angles, VV_converged_timestep,
                                                  normalization_VV)
                                                  
                    np.savetxt('{}/Position_VV.txt'.format(dir_run_feb_output), VV_position)
                    np.savetxt('{}/SimulatedLoads_VV.txt'.format(dir_run_feb_output), VV_simulated_loads)
                    np.savetxt('{}/SimulatedAngles_VV.txt'.format(dir_run_feb_output), VV_simulated_angles)
                    np.savetxt('{}/RobotAngles_VV.txt'.format(dir_run_feb_output), VV_robot_angles)                                                  

                    if VV_norm_error >= 3.4:
                        VV_norm_error = 3.4
                else:
                    VV_norm_error = 3.4
            else:
                VV_norm_error = 3.4
                
            if AP_check == 1:
                if len(AP_converged_times) > 0:
                    AP_converged_timestep = AP_converged_times[-1]
                    AP_simulated_loads = loads_applied(AP_converged_times, min_load_AP, max_load_AP)
                    AP_simulated_angles = AP_position[4, :]
                    AP_robot_angles = get_robot_position_AP(AP_simulated_loads)
                    AP_norm_error = calc_error_ap(AP_robot_angles, AP_simulated_angles, AP_converged_timestep,
                                                  normalization_AP)
                                                  
                    np.savetxt('{}/Position_AP.txt'.format(dir_run_feb_output), AP_position)
                    np.savetxt('{}/SimulatedLoads_AP.txt'.format(dir_run_feb_output), AP_simulated_loads)
                    np.savetxt('{}/SimulatedPosition_AP.txt'.format(dir_run_feb_output), AP_simulated_angles)
                    np.savetxt('{}/RobotPosition_AP.txt'.format(dir_run_feb_output), AP_robot_angles)                                                   

                    if AP_norm_error >= 3.4:
                        AP_norm_error = 3.4
                else:
                    AP_norm_error = 3.4
            else:
                AP_norm_error = 3.4
                
            if IE_check == 1:
                if len(IE_converged_times) > 0:
                    IE_converged_timestep = IE_converged_times[-1]
                    IE_simulated_loads = loads_applied(IE_converged_times, min_load_IE, max_load_IE)
                    IE_simulated_angles = IE_position[2, :]
                    IE_robot_angles = get_robot_angle_IE(IE_simulated_loads)
                    IE_norm_error = calc_error_ie(IE_robot_angles, IE_simulated_angles, IE_converged_timestep,
                                                  normalization_IE)
                                                  
                    np.savetxt('{}/Position_IE.txt'.format(dir_run_feb_output), IE_position)
                    np.savetxt('{}/SimulatedLoads_IE.txt'.format(dir_run_feb_output), IE_simulated_loads)
                    np.savetxt('{}/SimulatedAngles_IE.txt'.format(dir_run_feb_output), IE_simulated_angles)
                    np.savetxt('{}/RobotAngles_IE.txt'.format(dir_run_feb_output), IE_robot_angles)                                                        

                    if IE_norm_error >= 3.4:
                        IE_norm_error = 3.4
                else:
                    IE_norm_error = 3.4
            else:
                IE_norm_error = 3.4

        else:  # Axial load application failed
            if len(fl_converged_times_PS_F_AL) == 0:   
                VV_norm_error = 4
                AP_norm_error = 4
                IE_norm_error = 4
            else: 
                if last_convergedtime >= 0.8:
                    VV_norm_error = 3.5
                    AP_norm_error = 3.5
                    IE_norm_error = 3.5        
                elif last_convergedtime >= 0.6:
                    VV_norm_error = 3.6
                    AP_norm_error = 3.6
                    IE_norm_error = 3.6
                elif last_convergedtime >= 0.4:
                    VV_norm_error = 3.7
                    AP_norm_error = 3.7
                    IE_norm_error = 3.7   
                elif last_convergedtime >= 0.2:
                    VV_norm_error = 3.8
                    AP_norm_error = 3.8
                    IE_norm_error = 3.8     
                elif last_convergedtime >= 0:
                    VV_norm_error = 3.9
                    AP_norm_error = 3.9
                    IE_norm_error = 3.9     


    # TOTAL ERROR --------------------------------------------------------------------------------------------------
    total_error = (VV_norm_error + AP_norm_error + IE_norm_error) / 3

    # WRITE VALUES TO FILES --------------------------------------------------------------------------------------------
    with open('{}/prestretch_values_LCL_MCL_ACL_PCL.txt'.format(dir_run_feb_output), 'a') as pres:
        pres.write(
            'prestretch values: {} {} {} {} {} {} {} {} Time: {} \n'.format(str(PS_values[0]), str(PS_values[1]), str(PS_values[2]), str(PS_values[3]), str(YMVals[0]), str(YMVals[1]), str(YMVals[2]), str(YMVals[3]), datetime.now().strftime("%d/%m/%Y %H:%M:%S")))  # LCL, MCL, ACL, PCL

    with open('{}/prestretch_values_LCL_MCL_ACL_PCL_rounded.txt'.format(dir_run_feb_output), 'a') as rpres:
        rpres.write(
            'prestretch values: {} {} {} {} {} {} {} {} Time: {} \n'.format(str(prestretch_values[0]), str(prestretch_values[1]), str(prestretch_values[2]), str(prestretch_values[3]), str(YMValues[0]), str(YMValues[1]), str(YMValues[2]), str(YMValues[3]), datetime.now().strftime("%d/%m/%Y %H:%M:%S")))  # LCL, MCL, ACL, PCL

    with open('{}/counters.txt'.format(dir_run_feb_output), 'a') as cou:
        cou.write('run: {} VV: {} AP: {} IE: {} Time: {} \n'.format(str(counter_run), str(counter_VV), str(counter_AP),
                                                                    str(counter_IE),
                                                                    datetime.now().strftime("%d/%m/%Y %H:%M:%S")))

    with open('{}/{}_normalized_errors.txt'.format(dir_run_feb_output, model_nr), 'a+') as ne:
        ne.write('Total: {} VV: {} AP: {} IE: {} Time: {} \n'.format(total_error, VV_norm_error, AP_norm_error,
                                                                     IE_norm_error,
                                                                     datetime.now().strftime("%d/%m/%Y %H:%M:%S")))

    return total_error


# --- MAIN -------------------------------------------------------------------------------------------------------------
def main():
    if number == 1:
        ACL_init = float(0.95)
        PCL_init = float(1.00)
        MCL_init = float(1.00)
        LCL_init = float(1.00)

    elif number == 2:
        ACL_init = float(1.00)
        PCL_init = float(1.05)
        MCL_init = float(1.00)
        LCL_init = float(1.00)

    elif number == 3:
        ACL_init = float(1.00)
        PCL_init = float(1.00)
        MCL_init = float(1.05)
        LCL_init = float(1.00)

    elif number == 4:
        ACL_init = float(1.00)
        PCL_init = float(1.00)
        MCL_init = float(1.00)
        LCL_init = float(0.95)

    elif number == 5:
        ACL_init = float(1.00)
        PCL_init = float(1.00)
        MCL_init = float(1.00)
        LCL_init = float(1.00)
        
    ACL_Youngs = float(0.5)
    PCL_Youngs = float(0.5)
    MCL_Youngs = float(0.5)
    LCL_Youngs = float(0.5)
    

    ToOptimize = [LCL_init, MCL_init, ACL_init, PCL_init, ACL_Youngs, PCL_Youngs, MCL_Youngs, LCL_Youngs]

    if method == 'TNC':
        bounds = ((0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0, 1), (0, 1), (0, 1), (0, 1))

        minimizer_kwargs = {"method": "TNC", "bounds": bounds}
        results = basinhopping(objective_function, ToOptimize, minimizer_kwargs=minimizer_kwargs)
        
    elif method == 'LBFGSB':
        bounds = ((0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0, 1), (0, 1), (0, 1), (0, 1))

        minimizer_kwargs = {"method": "L-BFGS-B", "bounds": bounds}
        results = basinhopping(objective_function, ToOptimize, minimizer_kwargs=minimizer_kwargs)

if __name__ == "__main__":
    main()
