import numpy as np
import thread
import xml.etree.cElementTree as ET
import subprocess
import time
from scipy.optimize import basinhopping
from datetime import datetime
from sympy import Point3D, Line3D
import matplotlib.pyplot as plt
import os

model_nr = 'oks007'

method = 'TNC'
# method = 'LBFGSB'

# number = 1
# number = 2
# number = 3
# number = 4
number = 5

global counter_AP
global counter_IE
global counter_VV
global counter_run

counter_AP = 0
counter_IE = 0
counter_VV = 0
counter_run = 0
thread_count = 0

# ----------------------------------------------------------------------------------------------------------------------
# --- SET DIRECTORIES AND FILEPATHS ------------------------------------------------------------------------------------
dir_robot_data = r'./Robot_data'
dir_template_feb = r'./Templates_{}_{}_{}'.format(model_nr, method, number)
dir_run_feb = r'./Runfolder_{}_{}_{}'.format(model_nr, method, number)
dir_run_feb_output = r'{}/Output'.format(dir_run_feb)
dir_nodenrs = r'./Nodenrs ACS'

# file names
feb_name_imaging_posture = '{}'.format(model_nr) 
feb_name_PS_F_AL = '{}_PS_F_AL'.format(model_nr)    # Prestretch, flexion and axial load
feb_name_VV = '{}_VV'.format(model_nr)
feb_name_IE = '{}_IE'.format(model_nr)
feb_name_AP = '{}_AP'.format(model_nr)
file_names_list = [feb_name_PS_F_AL, feb_name_VV, feb_name_IE, feb_name_AP]

log_file_PS_F_AL = ('{}/{}.log'.format(dir_run_feb, feb_name_PS_F_AL))
log_file_VV = ('{}/{}.log'.format(dir_run_feb, feb_name_VV))
log_file_AP = ('{}/{}.log'.format(dir_run_feb, feb_name_AP))
log_file_IE = ('{}/{}.log'.format(dir_run_feb, feb_name_IE))

min_load_VV = -5000
max_load_VV = 5000
min_load_AP = -80
max_load_AP = 60
min_load_IE = -4000
max_load_IE = 4000

# Node numbers used in ACS
node_numbers_fem = np.loadtxt(r'{}/{}_fem_alm_nodes.txt'.format(dir_nodenrs, model_nr))
node_numbers_tib = np.loadtxt(r'{}/{}_tib_alm_nodes.txt'.format(dir_nodenrs, model_nr))

########################################################################################################################
# ------------------------------------------------- LOAD ROBOT DATA ----------------------------------------------------
# Load robot data look up tables
robot_AP_forces = np.load(r'{}/{}/AP_forces.npy'.format(dir_robot_data, model_nr))
robot_AP_positions = np.load(r'{}/{}/AP_position.npy'.format(dir_robot_data, model_nr))
robot_IE_moments = np.load(r'{}/{}/IE_moments.npy'.format(dir_robot_data, model_nr))
robot_IE_angles = np.load(r'{}/{}/IE_angles.npy'.format(dir_robot_data, model_nr))
robot_VV_moments = np.load(r'{}/{}/VV_moments.npy'.format(dir_robot_data, model_nr))
robot_VV_angles = np.load(r'{}/{}/VV_angles.npy'.format(dir_robot_data, model_nr))

# Min & Max values Robot data
# min_robot_AP = np.load(r'{}/{}/AP_min_position.npy'.format(dir_robot_data, model_nr))
# min_robot_IE = np.load(r'{}/{}/IE_min_angle.npy'.format(dir_robot_data, model_nr))
# min_robot_VV = np.load(r'{}/{}/VV_min_angle.npy'.format(dir_robot_data, model_nr))
# max_robot_AP = np.load(r'{}/{}/AP_max_position.npy'.format(dir_robot_data, model_nr))
# max_robot_IE = np.load(r'{}/{}/IE_max_angle.npy'.format(dir_robot_data, model_nr))
# max_robot_VV = np.load(r'{}/{}/VV_max_angle.npy'.format(dir_robot_data, model_nr))
# normalization_AP = max([abs(min_robot_AP), abs(max_robot_AP)])
# normalization_IE = max([abs(min_robot_IE), abs(max_robot_IE)])
# normalization_VV = max([abs(min_robot_VV), abs(max_robot_VV)])


def find_closest_value_in_array(array, value):
    closest_value = array[np.abs(array - value).argmin()]
    index = list(array).index(closest_value)
    return index
    

def get_robot_position_AP(simulated_force_ap):
    if isinstance(simulated_force_ap, int):
        index_closest_value_AP_force = find_closest_value_in_array(robot_AP_forces, simulated_force_ap)
        AP_robot_positions = robot_AP_positions[index_closest_value_AP_force]
    
    else: 
        AP_robot_positions = []
        for i in range(0, len(simulated_force_ap)):
            index_closest_value_AP_force = find_closest_value_in_array(robot_AP_forces, simulated_force_ap[i])
            AP_robot_positions.append(robot_AP_positions[index_closest_value_AP_force])

    return AP_robot_positions


def get_robot_angle_IE(simulated_moment_ie):
    if isinstance(simulated_moment_ie, int):
        index_closest_value_IE_moment = find_closest_value_in_array(robot_IE_moments, simulated_moment_ie)
        IE_robot_angles = robot_IE_angles[index_closest_value_IE_moment]
    
    else:
        IE_robot_angles = []
        for i in range(0, len(simulated_moment_ie)):
            index_closest_value_IE_moment = find_closest_value_in_array(robot_IE_moments, simulated_moment_ie[i])
            IE_robot_angles.append(robot_IE_angles[index_closest_value_IE_moment])

    return IE_robot_angles


def get_robot_angle_VV(simulated_moment_vv):
    if isinstance(simulated_moment_vv, int):
        index_closest_value_VV_moment = find_closest_value_in_array(robot_VV_moments, simulated_moment_vv)
        VV_robot_angles = robot_VV_angles[index_closest_value_VV_moment]
    
    else:
        VV_robot_angles = []
        for i in range(0, len(simulated_moment_vv)):
            index_closest_value_VV_moment = find_closest_value_in_array(robot_VV_moments, simulated_moment_vv[i])
            VV_robot_angles.append(robot_VV_angles[index_closest_value_VV_moment])

    return VV_robot_angles
    

max_robot_AP = get_robot_position_AP(max_load_AP)
min_robot_AP = get_robot_position_AP(min_load_AP)
max_robot_IE = get_robot_angle_IE(max_load_IE)
min_robot_IE = get_robot_angle_IE(min_load_IE)
max_robot_VV = get_robot_angle_VV(max_load_VV)
min_robot_VV = get_robot_angle_VV(min_load_VV)
normalization_AP = max([abs(min_robot_AP), abs(max_robot_AP)])
normalization_IE = max([abs(min_robot_IE), abs(max_robot_IE)])
normalization_VV = max([abs(min_robot_VV), abs(max_robot_VV)])

########################################################################################################################
# ------------------------------------------------- FUNCTIONS ----------------------------------------------------------


def vec_mag(x):
    mag_x = np.sqrt(x[0] ** 2 + x[1] ** 2 + x[2] ** 2)

    return mag_x


def unit_vec(x, y):
    x = np.array(x)
    y = np.array(y)
    xy = y - x
    mag_xy = np.sqrt(xy[0] ** 2 + xy[1] ** 2 + xy[2] ** 2)
    u_xy = xy / mag_xy

    return mag_xy, u_xy


def set_load_values_in_lc(tree_knee, root_knee, max_load, min_load, simulation):
    if simulation == 'VV':
        lc = '9'
        max_time = 1.4
        min_time = 2.2
    elif simulation == 'AP':
        lc = '10'
        max_time = 1.4
        min_time = 2.2
    elif simulation == 'IE':
        lc = '7'
        max_time = 1.4
        min_time = 2.2

    for child in root_knee.findall('LoadData/loadcurve'):
        if child.attrib['id'] == '{}'.format(lc):
            max = '{},{}'.format(max_time, max_load)
            min = '{},{}'.format(min_time, min_load)
            child[2].text = max
            child[3].text = min

    return tree_knee


def set_prestretch_values(feb_file_name, tree_knee_feb, root_knee_feb, stretch_values, YMvalues):
    for child in root_knee_feb.findall('LoadData/loadcurve'):   
        if child.attrib['id'] == '4':                       # LCL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[0])
            step2 = '4,{}'.format(stretch_values[0])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2
             
        if child.attrib['id'] == '3':                       # MCL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[1])
            step2 = '4,{}'.format(stretch_values[1])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2
            
        if child.attrib['id'] == '1':                       # ACL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[2])
            step2 = '4,{}'.format(stretch_values[2])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2

        if child.attrib['id'] == '2':                       # PCL
            step0 = '0,1'
            step1 = '0.1,{}'.format(stretch_values[3])
            step2 = '4,{}'.format(stretch_values[3])
            child[0].text = step0
            child[1].text = step1
            child[2].text = step2
            
    for child in root_knee_feb.findall('Material/material'):
        if child.attrib['name'] == 'ACL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[0]))

        if child.attrib['name'] == 'PCL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[1]))

        if child.attrib['name'] == 'MCL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[2]))

        if child.attrib['name'] == 'LCL':
            for child1 in child:
                if child1.tag == 'elastic':
                    for child2 in child1:
                        if child2.tag == 'E':
                            child2.text = str('{}'.format(YMvalues[3]))
            

    tree_knee_feb.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_PS_F_AL), encoding="ISO-8859-1",
                        xml_declaration=True)


def run_feb(feb_file_name):
    global thread_count
    print('{} run starting'.format(feb_file_name))

    subprocess.call(["./FEBio-3.0.0/bin/febio3", '-i', "{}/{}.feb".format(dir_run_feb, feb_file_name), '-dump', ])

    print('{} run finished'.format(feb_file_name))
    thread_count += 1


def run_feb_from_restart(feb_file_name):
    global thread_count
    print('{} run starting'.format(feb_file_name))

    subprocess.call(["./FEBio-3.0.0/bin/febio3", '-r', "{}/{}.feb".format(dir_run_feb, feb_file_name)])

    print('{} run finished'.format(feb_file_name))
    thread_count += 1


def read_logfile(logfile):
    # OBTAIN ANATOMICAL LANDMARKS FROM .LOG FILE ---------------------------------------------------------------------------
    nodes_fem_1 = []
    nodes_fem_2 = []
    nodes_fem_3 = []
    nodes_fem_4 = []
    nodes_fem_5 = []
    nodes_fem_6 = []
    nodes_tib_1 = []
    nodes_tib_2 = []
    nodes_tib_3 = []
    nodes_tib_4 = []
    nodes_tib_5 = []
    nodes_tib_6 = []
    RCJ1 = []
    RCJ2 = []
    RCJ3 = []
    converged_times = []

    nodenr_fem_1 = int(node_numbers_fem[0])
    nodenr_fem_2 = int(node_numbers_fem[1])
    nodenr_fem_3 = int(node_numbers_fem[2])
    nodenr_fem_4 = int(node_numbers_fem[3])
    nodenr_fem_5 = int(node_numbers_fem[4])
    nodenr_fem_6 = int(node_numbers_fem[5])

    nodenr_tib_1 = int(node_numbers_tib[0])
    nodenr_tib_2 = int(node_numbers_tib[1])
    nodenr_tib_3 = int(node_numbers_tib[2])
    nodenr_tib_4 = int(node_numbers_tib[3])
    nodenr_tib_5 = int(node_numbers_tib[4])
    nodenr_tib_6 = int(node_numbers_tib[5])

    # READ LOG FILE
    with open(logfile, 'r') as inf:
        input_data = inf.readlines()
        for ii in range(0, len(input_data)):
            line = input_data[ii]

            if line.startswith('{} '.format(str(nodenr_fem_1))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_fem_1.append(xyz)

            if line.startswith('{} '.format(str(nodenr_fem_2))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_fem_2.append(xyz)

            if line.startswith('{} '.format(str(nodenr_fem_3))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_fem_3.append(xyz)

            if line.startswith('{} '.format(str(nodenr_fem_4))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_fem_4.append(xyz)

            if line.startswith('{} '.format(str(nodenr_fem_5))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_fem_5.append(xyz)

            if line.startswith('{} '.format(str(nodenr_fem_6))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_fem_6.append(xyz)

            if line.startswith('{} '.format(str(nodenr_tib_1))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_tib_1.append(xyz)

            if line.startswith('{} '.format(str(nodenr_tib_2))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_tib_2.append(xyz)

            if line.startswith('{} '.format(str(nodenr_tib_3))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_tib_3.append(xyz)

            if line.startswith('{} '.format(str(nodenr_tib_4))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_tib_4.append(xyz)

            if line.startswith('{} '.format(str(nodenr_tib_5))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_tib_5.append(xyz)

            if line.startswith('{} '.format(str(nodenr_tib_6))):
                xyz = [float(x) for x in input_data[ii].split(' ')]
                nodes_tib_6.append(xyz)

            elif line.startswith("Data = RCFx;RCFy;RCFz;RCMx;RCMy;RCMz\n"):
                xyz1 = [float(x) for x in input_data[ii + 1].split(',')][1:]
                RCJ1.append(xyz1)
                xyz2 = [float(x) for x in input_data[ii + 2].split(',')][1:]
                RCJ2.append(xyz2)
                xyz3 = [float(x) for x in input_data[ii + 3].split(',')][1:]
                RCJ3.append(xyz3)

            elif line.startswith("------- converged at time :"):
                A = input_data[ii][27:]
                B = A.replace('\n', '')
                converged_times.append(float(B))

    RCJ1 = np.asarray(RCJ1)
    RCJ2 = np.asarray(RCJ2)
    RCJ3 = np.asarray(RCJ3)

    list_of_nodes = [nodes_fem_1, nodes_fem_2, nodes_fem_3, nodes_fem_4, nodes_fem_5, nodes_fem_6,
                     nodes_tib_1, nodes_tib_2, nodes_tib_3, nodes_tib_4, nodes_tib_5, nodes_tib_6]

    return list_of_nodes, RCJ1, RCJ2, RCJ3, converged_times
    

def Knee_ACS(list_nodes, step):
    F1 = np.asarray(list_nodes[0][step][1:])           # Lateral femoral condyle   0
    F2 = np.asarray(list_nodes[1][step][1:])           # Medial femoral condyle    1
    F3 = np.asarray(list_nodes[2][step][1:])           # Proximal femur point 1    2
    F4 = np.asarray(list_nodes[3][step][1:])           # Proximal femur point 2    3
    F5 = np.asarray(list_nodes[4][step][1:])           # Proximal femur point 3    4
    F6 = np.asarray(list_nodes[5][step][1:])           # Proximal femur point 4    5

    T1 = np.asarray(list_nodes[6][step][1:])           # Most lateral point on the tibial plateau  6
    T2 = np.asarray(list_nodes[7][step][1:])           # Most medial point on the tibial plateau   7
    T3 = np.asarray(list_nodes[8][step][1:])           # Distal tibia point 1                      8
    T4 = np.asarray(list_nodes[9][step][1:])           # Distal tibia point 2                      9
    T5 = np.asarray(list_nodes[10][step][1:])           # Distal tibia point 3                     10
    T6 = np.asarray(list_nodes[11][step][1:])           # Distal tibia point 4                     11

    # Tibia local coordinate system
    tib_origin = (T1 + T2) / 2
    T_distal = (T3 + T4 + T5 + T6) / 4
    tib_z = (tib_origin - T_distal) / vec_mag(tib_origin - T_distal)        # z axis of tibia pointing superiorly
    tib_x_temporary = (T2 - T1) / vec_mag(T2 - T1)                          # temp. x axis pointing medially
    tib_y = np.cross(tib_z, tib_x_temporary)/vec_mag(np.cross(tib_z, tib_x_temporary))  # pointed posteriorly
    tib_x = np.cross(tib_y, tib_z)/ vec_mag(np.cross(tib_y, tib_z))         # pointed medially
    tib_base_vecs = np.array([tib_x, tib_y, tib_z])

    # Femur local coordinate system
    fem_origin = (F1 + F2) / 2
    F_proximal = (F3 + F4 + F5 + F6) / 4
    fem_X = (F2 - F1) / vec_mag(F2 - F1)                                                    # pointed medially
    fem_Z_temporary = (F_proximal - fem_origin) / vec_mag(F_proximal - fem_origin)
    fem_Y = np.cross(fem_Z_temporary, fem_X) / vec_mag(np.cross(fem_Z_temporary, fem_X))    # pointed posteriorly
    fem_Z = np.cross(fem_X, fem_Y) / vec_mag(np.cross(fem_X, fem_Y))                        # pointed superiorly
    fem_base_vecs = np.array([fem_X, fem_Y, fem_Z])

    return [fem_base_vecs, tib_base_vecs, fem_origin, tib_origin]

    
def TF_joint_transformation(femur_base_vecs, tibia_base_vecs, femur_origin, tibia_origin):
    # Changed the directions of the vectors to match Grood & Suntay.
    if model_nr == 'oks003':  # Left knee
        I = femur_base_vecs[0]
        J = femur_base_vecs[1]
        K = femur_base_vecs[2]
        i = tibia_base_vecs[0]
        j = tibia_base_vecs[1]
        k = tibia_base_vecs[2]
    else:                       # Right knee
        I = -1 * femur_base_vecs[0]
        J = -1 * femur_base_vecs[1]
        K = femur_base_vecs[2]
        i = -1 * tibia_base_vecs[0]
        j = -1 * tibia_base_vecs[1]
        k = tibia_base_vecs[2]

    e1 = I
    e3 = k
    e2 = np.cross(e3, e1) / vec_mag(np.cross(e3, e1))  # Floating axis
    floating_axis = e2

    alpha = np.arcsin(np.dot(-e2, K))  # Flexion (+) - Extension (-)
    flexion = alpha
    beta = np.arccos(np.dot(I, k))
    if model_nr == 'oks003':
        abduction = (np.pi / 2) - beta  #                                                       Left knee
        gamma = np.arcsin(np.dot(e2, i))  # External rotation (+) - Internal (-)                Left knee
    else:
        abduction = beta - (np.pi / 2)  # Abduction (+)(Valgus) - Adduction (-) (Varus)         Right knee
        gamma = np.arcsin(np.dot(-e2, i))  # External rotation (+) - Internal (-)               Right knee
    ext_rot = gamma

    # H: translation between femur and tibia vector, directed from femoral origin to tibial origin.
    H = tibia_origin - femur_origin
    q1 = np.dot(H, e1)  # Medial (-) - lateral (+) tibial thrust
    Lat_trans = q1
    q2 = np.dot(H, e2)  # Anterior (+) - posterior (-) tibial drawer
    A_trans = q2
    q3 = np.dot(-H, e3)  # Joint distraction (+) - compression (-): height of the femoral origin above the tibial transverse plane.
    Joint_distraction = q3

    return [flexion, abduction, ext_rot, Lat_trans, A_trans, Joint_distraction]


def get_kinematics_and_loads(log_file):
    list_of_all_nodes, rcj1, rcj2, rcj3, converged_times = read_logfile(log_file)

    all_steps_ACS = []
    all_steps_rot_trans = []
    rot_trans_from_init = []

    for n in range(0, len(converged_times)):
        init_ACS = Knee_ACS(list_of_all_nodes, step=0)
        init_rot_trans = TF_joint_transformation(init_ACS[0], init_ACS[1], init_ACS[2], init_ACS[3])

        all_steps_ACS.append(Knee_ACS(list_of_all_nodes, step=n))
        all_steps_rot_trans.append(TF_joint_transformation(all_steps_ACS[n][0], all_steps_ACS[n][1], all_steps_ACS[n][2], all_steps_ACS[n][3]))
        rot_trans_from_init.append(np.array(all_steps_rot_trans[n]) - np.array(init_rot_trans))

    position = np.transpose(all_steps_rot_trans)

    return position, rcj1, rcj2, rcj3, converged_times


def loads_applied(converged_times, min_load, max_load):
    applied_loads = []
    for i in range(0, len(converged_times)):
        if converged_times[i] <= 1.0:
            load = 0

        elif converged_times[i] > 1.0 and converged_times[i] <= 1.4:
            b = max_load - ((max_load - 0) / (1.4 - 1.0)) * 1.4
            load = (((max_load - 0) / (1.4 - 1.0)) * converged_times[i]) + b

        elif converged_times[i] > 1.4 and converged_times[i] <= 2.2:
            b = min_load - (((min_load - max_load) / (2.2 - 1.4)) * 2.2)
            load = (((min_load - max_load) / (2.2 - 1.4)) * converged_times[i]) + b

        applied_loads.append(load)

    return applied_loads


def calc_error_vv(robot_data, simulated_data, converged_timestep, normalization_value):
    global counter_VV
    times = [1.0, 1.3, 1.6, 1.9, 2.2]

    if converged_timestep >= times[4]:
        error = np.sqrt(
            (np.array([x ** 2 for x in np.subtract(simulated_data, robot_data)]).sum()) / len(simulated_data))
        error_norm = error / normalization_value
        counter_VV = 0

    else:
        if converged_timestep >= times[3]:
            error_norm = 3.125
        elif converged_timestep >= times[2]:
            error_norm = 3.150
        elif converged_timestep >= times[1]:
            error_norm = 3.175
        elif converged_timestep >= times[0]:
            error_norm = 3.200

        counter_VV += 1
        if counter_VV >= 100:
            counter_VV_100 = 100
            error_norm = error_norm * (1 + (0.001 * (counter_VV_100 - 1)))
        else:
            error_norm = error_norm * (1 + (0.001 * (counter_VV - 1)))

    return error_norm


def calc_error_ap(robot_data, simulated_data, converged_timestep, normalization_value):
    global counter_AP
    times = [1.0, 1.3, 1.6, 1.9, 2.2]

    if converged_timestep >= times[4]:
        error = np.sqrt(
            (np.array([x ** 2 for x in np.subtract(simulated_data, robot_data)]).sum()) / len(simulated_data))
        error_norm = error / normalization_value
        counter_AP = 0

    else:
        if converged_timestep >= times[3]:
            error_norm = 3.125
        elif converged_timestep >= times[2]:
            error_norm = 3.150
        elif converged_timestep >= times[1]:
            error_norm = 3.175
        elif converged_timestep >= times[0]:
            error_norm = 3.200

        counter_AP += 1
        if counter_AP >= 100:
            counter_AP_100 = 100
            error_norm = error_norm * (1 + (0.001 * (counter_AP_100 - 1)))
        else:
            error_norm = error_norm * (1 + (0.001 * (counter_AP - 1)))

    return error_norm


def calc_error_ie(robot_data, simulated_data, converged_timestep, normalization_value):
    global counter_IE
    times = [1.0, 1.3, 1.6, 1.9, 2.2]

    if converged_timestep >= times[4]:
        error = np.sqrt(
            (np.array([x ** 2 for x in np.subtract(simulated_data, robot_data)]).sum()) / len(simulated_data))
        error_norm = error / normalization_value
        counter_IE = 0

    else:
        if converged_timestep >= times[3]:
            error_norm = 3.125
        elif converged_timestep >= times[2]:
            error_norm = 3.150
        elif converged_timestep >= times[1]:
            error_norm = 3.175
        elif converged_timestep >= times[0]:
            error_norm = 3.200

        counter_IE += 1
        if counter_IE >= 100:
            counter_IE_100 = 100
            error_norm = error_norm * (1 + (0.001 * (counter_IE_100 - 1)))
        else:
            error_norm = error_norm * (1 + (0.001 * (counter_IE - 1)))

    return error_norm


def find_closest_value_in_array(array, value):
    closest_value = array[np.abs(array - value).argmin()]
    index = list(array).index(closest_value)
    return index
    

def axialload_application_check(log_file):
    converged_times = []

    with open(log_file, 'r') as inf:
        input_data = inf.readlines()
        for ii in range(0, len(input_data)):
            line = input_data[ii]

            if line.startswith("------- converged at time :"):
                A = input_data[ii][27:]
                B = A.replace('\n', '')
                converged_times.append(float(B))

    if converged_times[-1] == 1.0:
        axialload_done = 1
    else:
        axialload_done = 0

    last_convergedtime = converged_times[-1]

    return axialload_done, last_convergedtime


def check_if_combination_already_ran(PS, YMV):
    AP_error_prev = np.nan
    VV_error_prev = np.nan
    IE_error_prev = np.nan
    already_ran = 0
    # prestretch_values = [LCL, MCL, ACL, PCL]

    if counter_run > 0:
        prestretch_values_file = '{}/prestretch_values_LCL_MCL_ACL_PCL_rounded.txt'.format(dir_run_feb_output)
        total_errors_file = '{}/{}_normalized_errors.txt'.format(dir_run_feb_output, model_nr)

        with open(prestretch_values_file, 'r') as inf:
            input_data = inf.readlines()

            values = []
            for ii in range(0, len(input_data)):
                values.append([x for x in input_data[ii].split(' ')])

            inf.close()

        ACL = []
        PCL = []
        MCL = []
        LCL = []
        ACL_Y = []
        PCL_Y = []
        MCL_Y = []
        LCL_Y = []
        
        for i in range(0, len(values)):
            ACL.append(float(values[i][4]))
            PCL.append(float(values[i][5]))
            MCL.append(float(values[i][3]))
            LCL.append(float(values[i][2]))
            ACL_Y.append(float(values[i][6]))
            PCL_Y.append(float(values[i][7]))
            MCL_Y.append(float(values[i][8]))
            LCL_Y.append(float(values[i][9]))

        ran = []
        for n in range(0, len(ACL)):
            if ACL[n] == PS[2] and PCL[n] == PS[3] and MCL[n] == PS[1] and LCL[n] == PS[0] and ACL_Y[n] == YMV[0] and PCL_Y[n] == YMV[1] and MCL_Y[n] == YMV[2] and LCL_Y[n] == YMV[3]:
                ran.append(1)
            else:
                ran.append(0)

        if (1 in ran):
            already_ran = 1

            with open(total_errors_file, 'r') as inf:
                input_data = inf.readlines()

                values = []
                for ii in range(0, len(input_data)):
                    values.append([x for x in input_data[ii].split(' ')])

                inf.close()

            AP_error = []
            VV_error = []
            IE_error = []

            for k in range(0, len(values)):
                AP_error.append(float(values[k][5]))
                VV_error.append(float(values[k][3]))
                IE_error.append(float(values[k][7]))

            for m in range(0, len(ran)):
                if ran[m] == 1:
                    AP_error_prev = AP_error[m]
                    VV_error_prev = VV_error[m]
                    IE_error_prev = IE_error[m]

        else:
            already_ran = 0

    return already_ran, AP_error_prev, VV_error_prev, IE_error_prev


########################################################################################################################
# ----------------------------------------------END OF FUNCITONS -------------------------------------------------------
########################################################################################################################

# ----------------------------------------------------------------------------------------------------------------------
# --- SET APPLIED LOAD LOADCURVES IN TEMPLATE SIMULATION FILES ---------------------------------------------------------
feb_file_VV = r'{}/{}.feb'.format(dir_template_feb, feb_name_VV)
tree_knee_VV = ET.parse(r'{}'.format(feb_file_VV))
root_knee_VV = tree_knee_VV.getroot()

feb_file_AP = r'{}/{}.feb'.format(dir_template_feb, feb_name_AP)
tree_knee_AP = ET.parse(r'{}'.format(feb_file_AP))
root_knee_AP = tree_knee_AP.getroot()

feb_file_IE = r'{}/{}.feb'.format(dir_template_feb, feb_name_IE)
tree_knee_IE = ET.parse(r'{}'.format(feb_file_IE))
root_knee_IE = tree_knee_IE.getroot()

tree_knee_VV_updated = set_load_values_in_lc(tree_knee_VV, root_knee_VV, max_load_VV, min_load_VV, 'VV')
tree_knee_VV_updated.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_VV), encoding="ISO-8859-1", xml_declaration=True)

tree_knee_AP_updated = set_load_values_in_lc(tree_knee_AP, root_knee_AP, max_load_AP, min_load_AP, 'AP')
tree_knee_AP_updated.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_AP), encoding="ISO-8859-1", xml_declaration=True)

tree_knee_IE_updated = set_load_values_in_lc(tree_knee_IE, root_knee_IE, max_load_IE, min_load_IE, 'IE')
tree_knee_IE_updated.write(r'{}/{}.feb'.format(dir_run_feb, feb_name_IE), encoding="ISO-8859-1", xml_declaration=True)

# ----------------------------------------------------------------------------------------------------------------------
# --- OBJECTIVE FUNCTION -----------------------------------------------------------------------------------------------
def objective_function(toOptimize):
    global thread_count
    global counter_run
    global file_names_list
    global total_error
    global VV_norm_error
    global AP_norm_error
    global IE_norm_error
    
    PS_values = toOptimize[0:4]     # PS_values = [LCL_init, MCL_init, ACL_init, PCL_init]
    
    ACL_YMvalues = [73, 83, 93, 103, 113, 123, 133, 143, 153, 163, 173]
    PCL_YMvalues = [118, 128, 138, 148, 158, 168, 178, 188, 198, 208, 218]
    MCL_YMvalues = [174, 184, 194, 204, 214, 224, 234, 244, 254, 264, 274]
    LCL_YMvalues = [230, 240, 250, 260, 270, 280, 290, 300, 310, 320, 330]
    
    # YMVals = toOptimize[4:8]        # ACL, PCL, MCL, LCL
    YMVals = [ACL_YMvalues[int(round(toOptimize[4]*10))], PCL_YMvalues[int(round(toOptimize[5]*10))], MCL_YMvalues[int(round(toOptimize[6]*10))], LCL_YMvalues[int(round(toOptimize[7]*10))]]
   
    ACL_PS = PS_values[2]
    PCL_PS = PS_values[3]
    MCL_PS = PS_values[1]
    LCL_PS = PS_values[0]

    ACL_PS_r = round(float(ACL_PS), 3)
    PCL_PS_r = round(float(PCL_PS), 3)
    MCL_PS_r = round(float(MCL_PS), 3)
    LCL_PS_r = round(float(LCL_PS), 3)

    LCL = float(LCL_PS_r)
    MCL = float(MCL_PS_r)
    ACL = float(ACL_PS_r)
    PCL = float(PCL_PS_r)

    prestretch_values = [LCL, MCL, ACL, PCL]
    
    ACL_YM = YMVals[0]
    PCL_YM = YMVals[1]
    MCL_YM = YMVals[2]
    LCL_YM = YMVals[3]
    
    ACL_YM_r = YMVals[0]
    PCL_YM_r = YMVals[1]
    MCL_YM_r = YMVals[2]
    LCL_YM_r = YMVals[3]
    
    ACL_Y = float(ACL_YM_r)
    PCL_Y = float(PCL_YM_r)
    MCL_Y = float(MCL_YM_r)
    LCL_Y = float(LCL_YM_r)
    
    YMValues = [ACL_Y, PCL_Y, MCL_Y, LCL_Y] 

    already_ran, previousAPerror, previousVVerror, previousIEerror = check_if_combination_already_ran(prestretch_values, YMValues)

    if already_ran == 1:
        VV_norm_error = previousVVerror
        AP_norm_error = previousAPerror
        IE_norm_error = previousIEerror
        total_error = (VV_norm_error + AP_norm_error + IE_norm_error) / 3

    elif already_ran == 0:
        # DELETE .log, .dmp, .xplt files BEFORE NEXT RUN------------------------------------------------------------
        if counter_run > 0:
            for i in range(0, len(file_names_list)):
                if os.path.isfile('{}/{}.log'.format(dir_run_feb, file_names_list[i])) == True:
                    os.remove('{}/{}.log'.format(dir_run_feb, file_names_list[i]))
                if os.path.isfile('{}/{}.xplt'.format(dir_run_feb, file_names_list[i])) == True:
                    os.remove('{}/{}.xplt'.format(dir_run_feb, file_names_list[i]))
                if os.path.isfile('{}/{}.dmp'.format(dir_run_feb, file_names_list[i])) == True:
                    os.remove('{}/{}.dmp'.format(dir_run_feb, file_names_list[i]))
            if os.path.isfile('{}/{}.feb'.format(dir_run_feb, feb_name_PS_F_AL)) == True:
                os.remove('{}/{}.feb'.format(dir_run_feb, feb_name_PS_F_AL))

        counter_run += 1

        # SET PRESTRETCH IN IMAGING POSTURE FEB FILE & RUN ---------------------------------------------------------
        feb_file = r'{}/{}.feb'.format(dir_template_feb, feb_name_imaging_posture)
        tree_knee = ET.parse(r'{}'.format(feb_file))
        root_knee = tree_knee.getroot()
        set_prestretch_values(feb_name_imaging_posture, tree_knee, root_knee, prestretch_values, YMValues)

        run_feb(feb_name_PS_F_AL)
        flexion_position_PS_F_AL, fl_loads_rcj1_PS_F_AL, fl_loads_rcj2_PS_F_AL, fl_loads_rcj3_PS_F_AL, fl_converged_times_PS_F_AL = get_kinematics_and_loads(log_file_PS_F_AL)
            
        if len(fl_converged_times_PS_F_AL) == 0:
            axialload_succes = 0
        else:
            axialload_succes, last_convergedtime = axialload_application_check(log_file_PS_F_AL)

        if axialload_succes == 1:
            with open('{}/flexion_angle_PS_F_AL.txt'.format(dir_run_feb_output), 'a') as pres1:
                pres1.write('flexion angle: {} Time: {} \n'.format(str(flexion_position_PS_F_AL[0, -1]), datetime.now().strftime("%d/%m/%Y %H:%M:%S")))
            thread_count = 0
            num_method = 3

            thread.start_new_thread(run_feb_from_restart, ('{}'.format(feb_name_VV),))
            thread.start_new_thread(run_feb_from_restart, ('{}'.format(feb_name_AP),))
            thread.start_new_thread(run_feb_from_restart, ('{}'.format(feb_name_IE),))

            while thread_count < num_method:
                print(thread_count)
                time.sleep(1)

            VV_position, VV_loads_rcj1, VV_loads_rcj2, VV_loads_rcj3, VV_converged_times = get_kinematics_and_loads(
                log_file_VV)                    
            AP_position, AP_loads_rcj1, AP_loads_rcj2, AP_loads_rcj3, AP_converged_times = get_kinematics_and_loads(
                log_file_AP)
            IE_position, IE_loads_rcj1, IE_loads_rcj2, IE_loads_rcj3, IE_converged_times = get_kinematics_and_loads(
                log_file_IE)

            if len(VV_converged_times) == 0:
                VV_check = 0
            else:
                VV_check = 1
                
            if len(AP_converged_times) == 0:
                AP_check = 0
            else:
                AP_check = 1
                
            if len(IE_converged_times) == 0:
                IE_check = 0
            else:
                IE_check = 1

            # CALCULATE VV ERROR ---------------- Position: 0: FE, 1: VV, 2: IE, 3: ML, 4: AP, 5: SI ---------------
            if VV_check == 1:
                if len(VV_converged_times) > 0:
                    VV_converged_timestep = VV_converged_times[-1]
                    VV_simulated_loads = loads_applied(VV_converged_times, min_load_VV, max_load_VV)
                    VV_simulated_angles = VV_position[1, :]
                    VV_robot_angles = get_robot_angle_VV(VV_simulated_loads)
                    VV_norm_error = calc_error_vv(VV_robot_angles, VV_simulated_angles, VV_converged_timestep,
                                                  normalization_VV)
                                                  
                    np.savetxt('{}/Position_VV.txt'.format(dir_run_feb_output), VV_position)
                    np.savetxt('{}/SimulatedLoads_VV.txt'.format(dir_run_feb_output), VV_simulated_loads)
                    np.savetxt('{}/SimulatedAngles_VV.txt'.format(dir_run_feb_output), VV_simulated_angles)
                    np.savetxt('{}/RobotAngles_VV.txt'.format(dir_run_feb_output), VV_robot_angles)                                                  

                    if VV_norm_error >= 3.4:
                        VV_norm_error = 3.4
                else:
                    VV_norm_error = 3.4
            else:
                VV_norm_error = 3.4
                
            if AP_check == 1:
                if len(AP_converged_times) > 0:
                    AP_converged_timestep = AP_converged_times[-1]
                    AP_simulated_loads = loads_applied(AP_converged_times, min_load_AP, max_load_AP)
                    AP_simulated_angles = AP_position[4, :]
                    AP_robot_angles = get_robot_position_AP(AP_simulated_loads)
                    AP_norm_error = calc_error_ap(AP_robot_angles, AP_simulated_angles, AP_converged_timestep,
                                                  normalization_AP)
                                                  
                    np.savetxt('{}/Position_AP.txt'.format(dir_run_feb_output), AP_position)
                    np.savetxt('{}/SimulatedLoads_AP.txt'.format(dir_run_feb_output), AP_simulated_loads)
                    np.savetxt('{}/SimulatedPosition_AP.txt'.format(dir_run_feb_output), AP_simulated_angles)
                    np.savetxt('{}/RobotPosition_AP.txt'.format(dir_run_feb_output), AP_robot_angles)                                                   

                    if AP_norm_error >= 3.4:
                        AP_norm_error = 3.4
                else:
                    AP_norm_error = 3.4
            else:
                AP_norm_error = 3.4
                
            if IE_check == 1:
                if len(IE_converged_times) > 0:
                    IE_converged_timestep = IE_converged_times[-1]
                    IE_simulated_loads = loads_applied(IE_converged_times, min_load_IE, max_load_IE)
                    IE_simulated_angles = IE_position[2, :]
                    IE_robot_angles = get_robot_angle_IE(IE_simulated_loads)
                    IE_norm_error = calc_error_ie(IE_robot_angles, IE_simulated_angles, IE_converged_timestep,
                                                  normalization_IE)
                                                  
                    np.savetxt('{}/Position_IE.txt'.format(dir_run_feb_output), IE_position)
                    np.savetxt('{}/SimulatedLoads_IE.txt'.format(dir_run_feb_output), IE_simulated_loads)
                    np.savetxt('{}/SimulatedAngles_IE.txt'.format(dir_run_feb_output), IE_simulated_angles)
                    np.savetxt('{}/RobotAngles_IE.txt'.format(dir_run_feb_output), IE_robot_angles)                                                        

                    if IE_norm_error >= 3.4:
                        IE_norm_error = 3.4
                else:
                    IE_norm_error = 3.4
            else:
                IE_norm_error = 3.4

        else:  # Axial load application failed
            if len(fl_converged_times_PS_F_AL) == 0:   
                VV_norm_error = 4
                AP_norm_error = 4
                IE_norm_error = 4
            else: 
                if last_convergedtime >= 0.8:
                    VV_norm_error = 3.5
                    AP_norm_error = 3.5
                    IE_norm_error = 3.5        
                elif last_convergedtime >= 0.6:
                    VV_norm_error = 3.6
                    AP_norm_error = 3.6
                    IE_norm_error = 3.6
                elif last_convergedtime >= 0.4:
                    VV_norm_error = 3.7
                    AP_norm_error = 3.7
                    IE_norm_error = 3.7   
                elif last_convergedtime >= 0.2:
                    VV_norm_error = 3.8
                    AP_norm_error = 3.8
                    IE_norm_error = 3.8     
                elif last_convergedtime >= 0:
                    VV_norm_error = 3.9
                    AP_norm_error = 3.9
                    IE_norm_error = 3.9      


    # TOTAL ERROR --------------------------------------------------------------------------------------------------
    total_error = (VV_norm_error + AP_norm_error + IE_norm_error) / 3

    # WRITE VALUES TO FILES --------------------------------------------------------------------------------------------
    with open('{}/prestretch_values_LCL_MCL_ACL_PCL.txt'.format(dir_run_feb_output), 'a') as pres:
        pres.write(
            'prestretch values: {} {} {} {} {} {} {} {} Time: {} \n'.format(str(PS_values[0]), str(PS_values[1]), str(PS_values[2]), str(PS_values[3]), str(YMVals[0]), str(YMVals[1]), str(YMVals[2]), str(YMVals[3]), datetime.now().strftime("%d/%m/%Y %H:%M:%S")))  # LCL, MCL, ACL, PCL

    with open('{}/prestretch_values_LCL_MCL_ACL_PCL_rounded.txt'.format(dir_run_feb_output), 'a') as rpres:
        rpres.write(
            'prestretch values: {} {} {} {} {} {} {} {} Time: {} \n'.format(str(prestretch_values[0]), str(prestretch_values[1]), str(prestretch_values[2]), str(prestretch_values[3]), str(YMValues[0]), str(YMValues[1]), str(YMValues[2]), str(YMValues[3]), datetime.now().strftime("%d/%m/%Y %H:%M:%S")))  # LCL, MCL, ACL, PCL

    with open('{}/counters.txt'.format(dir_run_feb_output), 'a') as cou:
        cou.write('run: {} VV: {} AP: {} IE: {} Time: {} \n'.format(str(counter_run), str(counter_VV), str(counter_AP),
                                                                    str(counter_IE),
                                                                    datetime.now().strftime("%d/%m/%Y %H:%M:%S")))

    with open('{}/{}_normalized_errors.txt'.format(dir_run_feb_output, model_nr), 'a+') as ne:
        ne.write('Total: {} VV: {} AP: {} IE: {} Time: {} \n'.format(total_error, VV_norm_error, AP_norm_error,
                                                                     IE_norm_error,
                                                                     datetime.now().strftime("%d/%m/%Y %H:%M:%S")))

    return total_error


# --- MAIN -------------------------------------------------------------------------------------------------------------
def main():
    if number == 1:
        ACL_init = float(0.95)
        PCL_init = float(1.00)
        MCL_init = float(1.00)
        LCL_init = float(1.00)

    elif number == 2:
        ACL_init = float(1.00)
        PCL_init = float(1.05)
        MCL_init = float(1.00)
        LCL_init = float(1.00)

    elif number == 3:
        ACL_init = float(1.00)
        PCL_init = float(1.00)
        MCL_init = float(1.05)
        LCL_init = float(1.00)

    elif number == 4:
        ACL_init = float(1.00)
        PCL_init = float(1.00)
        MCL_init = float(1.00)
        LCL_init = float(0.95)

    elif number == 5:
        ACL_init = float(1.00)
        PCL_init = float(1.00)
        MCL_init = float(1.00)
        LCL_init = float(1.00)
        
    ACL_Youngs = float(0.5)
    PCL_Youngs = float(0.5)
    MCL_Youngs = float(0.5)
    LCL_Youngs = float(0.5)
    

    ToOptimize = [LCL_init, MCL_init, ACL_init, PCL_init, ACL_Youngs, PCL_Youngs, MCL_Youngs, LCL_Youngs]

    if method == 'TNC':
        bounds = ((0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0, 1), (0, 1), (0, 1), (0, 1))

        minimizer_kwargs = {"method": "TNC", "bounds": bounds}
        results = basinhopping(objective_function, ToOptimize, minimizer_kwargs=minimizer_kwargs)
        
    elif method == 'LBFGSB':
        bounds = ((0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0.85, 1.05), (0, 1), (0, 1), (0, 1), (0, 1))

        minimizer_kwargs = {"method": "L-BFGS-B", "bounds": bounds}
        results = basinhopping(objective_function, ToOptimize, minimizer_kwargs=minimizer_kwargs)

if __name__ == "__main__":
    main()
